/*	May, June 1987	- modified for rapid read of database

	June 2, 1987 - added TFASTA
	March 30, 1988 - combined ffgetaa, fgetgb;
	April 8, 1988 - added PIRLIB format for unix
	Feb 4, 1989 - added universal subroutines for libraries
	December, 1995 - added range option file.name:1-1000

	copyright (c) 1987,1988,1989,1992,1995 William R. Pearson

	getnt.c	associated subroutines for matching sequences */

/*
8-April-88
	The compile time #define PIRLIB allows this routine to be used
	to read protein and DNA sequence libraries in the NBRF/PIR
	VAX/VMS library format.  That is:

	>P1;LCBO
	This is a line of description
	GTYH ... the sequence starts on this line

	This may ease conversion from UWGCG format libraries. It
	has not been extensively tested.

	In addition, sequence libraries with a '>' in the 4th position
	are recognized as NBRF format libraries for consistency with
	UWGCG

	February 4, 1988 - this starts a major revision of the getaa
	routines.  The goal is to be able to seach the following format
	libraries:

		0 - normal FASTA format
		1 - full Genbank tape format
		2 - NBRF/PIR CODATA format
		3 - EMBL/Swiss-prot format
		4 - Intelligentics format
		5 - NBRF/PIR VMS format
		6 - GCG 2bit format

		11 - NCBI setdb/blastp (1.3.2) AA

	see file altlib.h to confirm numbers

	This is done with a new global variable and a requirement for the
	FASTLIBS file.  The FASTLIBS file will now indicate both the sequence
	type (protein = 0, DNA = 1) and the file format (the numbers shown
	above, although intelligenetics may become an alternative to Pearson).
	This will be done by always using a function pointer for getlib and
	ranlib(), and setting up a bunch of different getlib() and ranlib()
	functions.  Openlib() will be substantially simplified.
*/

/* 	Nov 12, 1987	- this version checks to see if the sequence
	is DNA or protein by asking whether > 85% is A, C, G, T

	May 5, 1988 - modify the DNA/PROTEIN checker by re-reading
	DNA sequences in order to check for 'U'.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef __MWERKS__
#include <unix.h>
#include <stat.h>
#else
#include <sys/stat.h>
#endif

#include "uascii.gbl"
#include "ffasta.h"
#include "nxgetaa.h"
#include "jat.h"

#include "mpi.h"
int rank;
int numSlaves;
long endOffset;

#ifdef VMS
#define PIRLIB
#endif

#ifndef NOLIB
#include "altlib.h"
	extern short ldnaseq;
	long (*getlib)();
	long (*ranlib)();
#define GETLIB agetlib
#define RANLIB aranlib
#else
#define LASTLIB 10
#define GETLIB getlib
#define RANLIB ranlib
#endif

#define XTERNAL
#include "upam.gbl"
#undef XTERNAL

#define YES 1
#define NO 0
#define MAXLINE 512

#if (defined(__MWERKS__)  && __dest_os == __mac_os)
/* BIGBUFF allows the Mac to use setvbuf to set up a new buffer for reading
the library, that is 32K instead of 0.5k.  This appears to increase the speed
of the program by about 33%.  */
char *bigbuf=NULL;
long fbufsize=128L;

#define BIGBUF
extern char prompt[];
#include <StandardFile.h>
extern StandardFileReply freply;
/* extern short anvRef; */
extern FSSpec libSpec;
#endif

extern long sq0off;

/*=============================================================
*/
long getseq(char *filen, char *seq, long maxs, long *dnaseq)
{
	FILE  *fptr;
	char  line[512];
	char  *bp;
	long  i, j, n;
	long  ic;
	long  sstart, sstop, sset=0;

	

	sstart = sstop = -1;
#ifndef MSDOS
	if ((bp=strchr(filen,':'))!=NULL) {
#else
	if ((bp=strchr(filen+3,':'))!=NULL) {
#endif
	  *bp='\0';
	  if (*(bp+1)=='-') sscanf(bp+2,"%d",&sstop);
	  else sscanf(bp+1,"%d-%d",&sstart,&sstop);
	  sset=1;
	}

	// Try to open the file
	if (strncmp(filen,"-",1)!=0) {
    	if ((fptr=fopen(filen,"rb"))==NULL) {
      		fprintf(stderr," could not open \"%s\"\n",filen);
      		return 0;
    	}
  	}
  	else fptr = stdin;	

	if (sset==1) {
	  filen[strlen(filen)]=':';
	  if (sq0off==1 || sstart>1) sq0off = sstart;
	}

	// Read the sequence into the 'seq' array.
	n=0;
	while(my_fgets(line,sizeof(line),fptr)!=NULL) {
#ifdef PIRLIB
		if (line[0]=='>'&& (line[3]==';'||line[3]=='>'))
			my_fgets(line,sizeof(line),fptr);
		else
#endif
		if (line[0]!='>'&& line[0]!=';') {
			/* Translate the sequence from ASCII to amino acids 
			   in the order "ARNDCQEGHILKMFPSTWYVBZX"  */
		    for (i=0; (n<maxs)&&
			((ic=sascii[line[i]&AAMASK])<EL); i++)
				if (ic<NA) seq[n++]= ic;
		    if (ic == ES) break;
		    }
		}
	if (n==maxs) {
		fprintf(stderr," sequence may be truncated %d %d\n",n,maxs);
		fflush(stderr);
		}
	seq[n]= EOSEQ;

	/*  XXXXXXXXX What happens here if a short seq. such as AMGTC is input?  Would it
	    get converted to DNA ??? XXXXXXXXXXXXXX JAT 10/14/96 */
	    
	/* If it was not marked as a DNA sequence and it is greater than
	   85% A's, G's, C's, and T's then flag it as a DNA seq and reread it*/
	if (*dnaseq ==0 && (float)scanseq(seq,n,"ACGT")/(float)n > 0.85) {
		*dnaseq = 1;
				/* convert from protein to DNA sequence */
		sascii = nascii;
		fseek(fptr,0l,0);
		n=0;
		while(my_fgets(line,sizeof(line),fptr)!=NULL) {
#ifdef PIRLIB
			if (line[0]=='>'&& (line[3]==';'||line[3]=='>'))
				my_fgets(line,sizeof(line),fptr);
			else
#endif
			if (line[0]!='>'&& line[0]!=';') {
			    for (i=0; (n<maxs)&&
				((ic=sascii[line[i]&AAMASK])<EL); i++)
					if (ic<NA) seq[n++]= ic;
			    if (ic == ES) break;
			    }
			}
		if (n==maxs) {
			fprintf(stderr," sequence may be truncated %d %d\n",n,maxs);
			fflush(stderr);
			}
		seq[n]= EOSEQ;

		sq = nt;
		nsq = nnt;
		hsq = hnt;
		pam = npam;
		strcpy(sqnam,"nt");
		strcpy(sqtype,"DNA");
		}

	fclose(fptr);

	if ((sstart <= 0) || (sstop == -1)) {
	  if (sstart <= 0) sstart = 1;
	  if (sstop == -1) sstop = n;
	  sstart--;
	  sstop--;
	  for (i=0, j=sstart; j<=sstop; i++,j++)
	    seq[i] = seq[j];
	  n = sstop - sstart +1;
	  seq[n]=EOSEQ;
	}

	return n;
	}
/*=============================================================
*/
long gettitle(char *filen, char *title, long len)
{
	FILE *fptr;
	char line[512];
	char *bp;
	long ll,sset;
#ifdef MSDOS
	char *strpbrk();
#endif

#ifndef MSDOS
	if ((bp=strchr(filen,':'))!=NULL) { *bp='\0'; sset=1;}
#else
	if ((bp=strchr(filen+3,':'))!=NULL) { *bp='\0'; sset=1;}
#endif

	if ((fptr=fopen(filen,"rb"))==NULL) {
		fprintf(stderr," file %s was not found\n",filen);
		fflush(stderr);
		return 0;
		}

	if (sset==1) filen[strlen(filen)]=':';

	while(my_fgets(line,sizeof(line),fptr)!=0) {
		if (line[0]=='>'|| line[0]==';') goto found;
		}
	fclose(fptr);
	title[0]='\0';
	return 0;

found:
#ifdef PIRLIB
	if (line[0]=='>'&&(line[3]==';'||line[3]=='>')) {
		if ((bp = strchr(line,'\n'))!=NULL) *bp='\0';
		ll=strlen(line); line[ll++]=' '; line[ll]='\0';
		my_fgets(&line[ll],sizeof(line)-ll,fptr);
	}
#endif
#ifdef MSDOS
	bp = strpbrk(line,"\n\r");
#else
	bp = strchr(line,'\n');
#endif
	if (bp!=NULL) *bp = 0;
	strncpy(title,line,len);
	title[len-1]='\0';
	fclose(fptr);
	return strlen(title);
}	

#ifndef VMS
FILE *libf=NULL;
#else
long libf = -1;
#endif
#ifdef NOLIB
long leof = 0;
#endif

long lpos;
char lline[MAXLINE];
long lfflag=0;	/* flag for CRLF in EMBL CDROM files */
#define LFCHAR '\015'  /* for MWC 5.5 */


/*	the following is from fgetgb.c */

#if (defined(__MWERKS__)  && __dest_os == __mac_os)
SFTypeList llist={'TEXT',0L,0L,0L};
#define LLN 1
#endif

#include <fcntl.h>
#ifndef O_RAW
#ifdef O_BINARY
#define O_RAW O_BINARY
#else
#define O_RAW 0
#endif		/* O_BINARY */
#endif		/* O_RAW */
long libfd= -1;
#ifndef NOLIB
extern long deftype;	/* default library type */
extern long outtty;	/* flag for no interaction */
#ifndef UNIX
#define RBSTR "rb"	/* read file in binary mode */
#else
#define RBSTR "r"
#endif
#else
long deftype=0;
long outtty=1;
#endif
long libtype;		/* current open library type */
short sfnum=0;		/* superfamily number from types 0 and 5 */


/*=============================================================
*/
/* a file name for openlib may now include a library type suffix */
long openlib(char *lname, char *libenv)
{

	char rline[10],libn[120], *bp;
	long ftell();
	long wcnt, ll, opnflg;
        long  filelength;  
        struct stat     stat;
	long startingOffset;

	if (lname[0]=='#') return -9;
	wcnt = 0;

#ifndef NOLIB
	if (strlen(libenv)!=0) {
		strncpy(libn,libenv,sizeof(libn));
#ifdef UNIX
		strncat(libn,"/",sizeof(libn));
#endif
		strncat(libn,lname,sizeof(libn)-strlen(libn));
		}
	else strncpy(libn,lname,sizeof(libn));
#else
	strncpy(libn,lname,120);
#endif

	/* check for library type */
	if ((bp=strchr(libn,' '))!=NULL) {
	    *bp='\0';
	    sscanf(bp+1,"%d",&libtype);
	    if (libtype<0 || libtype >= LASTLIB) {
		fprintf(stderr," invalid library type: %d (>%d)- resetting\n%s\n",
			libtype,LASTLIB,lname);
		libtype=deftype;
		}
	    }
	else libtype=deftype;

#if (defined(__MWERKS__)  && __dest_os == __mac_os)
	HSetVol(NULL,libSpec.vRefNum,libSpec.parID);
	if (bigbuf==NULL) {
		if ((bp=getenv("BUFSIZE"))!=NULL) sscanf(bp,"%ld",&fbufsize);
		else fbufsize=128l;
		if ((bigbuf=malloc((long)(fbufsize*1024l)))==NULL)
			fprintf(stderr," cannot allocate %ld K buffer\n",fbufsize);
	}
#endif

#ifndef NOLIB
	getlib=getliba[libtype];
	ranlib=ranliba[libtype];

	if (libtype != INTELLIG)
	  sascii['0'] = sascii['1'] = sascii['2'] = NA;

l1:	if (libtype<=LASTTXT) opnflg=((libf=fopen(libn,"rb"))!=NULL);
#ifdef NCBIBL13
	else if (libtype==NCBIBL13) opnflg=(ncbl_openlib(libn)!= -1);
#endif

	if (!opnflg) {
#else
l1:	if ((libf=fopen(libn,"rb"))==NULL) {
#endif

#if (defined(__MWERKS__)  && __dest_os == __mac_os)
		rline[0]='\0';
		sprintf(prompt," cannot open %s\r Select library filename",libn);
		STFileDlog(prompt,&freply,llist,LLN);
		if (freply.sfGood==TRUE) {
			strcpy(libenv,"\0");	
			PtoCstr((StringPtr)freply.sfFile.name);
			strcpy(libn,(char *)freply.sfFile.name);
			strcpy(lname,libn);
/*			anvRef=freply.vRefNum;
			SetVol(NULL,anvRef);
*/
	  		libSpec.vRefNum = freply.sfFile.vRefNum;
			libSpec.parID = freply.sfFile.parID;
			HSetVol(NULL,libSpec.vRefNum,libSpec.parID);

			goto l1;
			}
		else return -1;
		}
#else
	if (outtty) {
	  fprintf(stderr," cannot open %s library\n",libn);
	  fprintf(stderr," enter new file name or <RET> to quit ");
	  fflush(stderr);
	  if (my_fgets(libn,120,stdin)==NULL) return -1;
	  if ((bp=strchr(libn,'\n'))!=0) *bp='\0';
	  if (strlen(libn)==0) return 0;
	  if (++wcnt > 10) return -1;
	  strcpy(lname,libn);
	  goto l1;
	}
	else return -1;
}
#endif	/* __MWERKS__ */

#ifndef NOLIB
#ifdef BIGBUF
	if (bigbuf!=NULL)
		setvbuf(libf,bigbuf,_IOFBF,(size_t)(fbufsize*1024l));
#endif
	if (libtype<=LASTTXT) {
		lpos = ftell(libf);
		if (my_fgets(lline,MAXLINE,libf)==NULL) return -1;
#if (defined(__MWERKS__)  && __dest_os == __mac_os)
		if (libtype==EMBLSWISS || libtype==VMSPIR ||libtype==FULLGB) {
			my_fgets(lline,MAXLINE,libf);
			lfflag = (lline[0]==LFCHAR);
			fseek(libf,0,0l);
			lpos = 0;
			my_fgets(lline,sizeof(lline)-1,libf);
			if (lfflag) {
				getc(libf);
		/*		fprintf(stderr," lfflag is set\n");  */
				}
			}
#endif
    }		
#else		/* NOLIB */
	lpos = ftell(libf);
	if (my_fgets(lline,MAXLINE,libf)==NULL) return -1;
	leof = 0;
#endif	/* NOLIB */


	MPI_Comm_rank( MPI_COMM_WORLD, &rank );
	MPI_Comm_size( MPI_COMM_WORLD, &numSlaves );  
	numSlaves -= 1; /* The master doesn't do any work */

        /* Get the file length */
        if(fstat(fileno(libf), &stat)) {
                printf("Couldn't get the database file length!\n");
		exit(0);
        }
        filelength = stat.st_size;
//	printf("%s filesize = %ld; numSlaves = %d\n", lname, filelength, numSlaves); // XXXXXXXXXXXXXXXX

	if (rank > 1) {
	    startingOffset = ((rank - 1)/numSlaves) * filelength;
	    if (fseek(libf, startingOffset, SEEK_SET)) {
		printf("fseek() failed to set offset at %ld!\n", startingOffset);
		printf("rank: %d  numslaves: %d  filelength: %ld\n", rank, numSlaves, filelength);
		MPI_Finalize();
		exit(1);
	    }
	    lpos = ftell(libf);
	    if (my_fgets(lline,MAXLINE,libf)==NULL) return -1;
	}
	endOffset = rank * filelength / numSlaves;

//	printf("startOffset = %ld; endOffset = %ld\n", startingOffset, endOffset); // XXXXXXXXXXXXXXXX

return 1;
}

/*=============================================================
*/
void closelib(void)
{
  if (libf!=NULL) {
    fclose(libf);
    libf = NULL;
  }
#ifndef NOLIB
#ifdef NCBIBL13
  if (libtype == NCBIBL13) ncbl_closelib();
#endif
#endif
}

/*=============================================================
*/
long agetlib(char *seq, long maxs, char *libstr, long *libpos, long *lcont)
{
	long ll;
	long ic;
	register char *cp, *seqp;
	register long *ap;
	char *seqm, *seqm1;
	char *linep, *bp;

	/* Don't read past our specified section of the database */
	if (ftell(libf) > endOffset && rank > 0) {
//	    printf("End of our section reached!\n");//XXXXXXXXXXXXXXXXX
	    return 0;
	}

	seqp = seq;
	seqm = &seq[maxs-9];
	seqm1 = seqm-1;
#if defined(TFASTA) || defined(TFASTX)
	ap = nascii;
#else
	ap = sascii;
#endif

	if (*lcont==0) {
#ifndef NOLIB
		while (lline[0]!='>' && lline[0]!=';') {
			lpos = ftell(libf);
			if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
		}
#ifdef SUPERFAMNUM
		if ((bp=strchr(lline,SFCHAR))!=NULL) {
			*bp='\0';
			sscanf(bp+1,"%d",&sfnum);
		      }
		else sfnum=0;
#else
		sfnum = 0;
#endif
		strncpy(libstr,lline + 1, 20);

 //       	printf("<%d> SEQ %s\n", rank, lline); /////////////////////////////////////////////
		if ((bp = strchr(libstr,' ')) != NULL) *bp = '\0';
		if ((bp = strchr(libstr,'\n')) != NULL) *bp = '\0';
		if ((bp = strchr(libstr,'\r')) != NULL) *bp = '\0';
		libstr[12]='\0';
		*libpos = lpos;
#else	/* NOLIB */
		if (leof) return 0;
		*libpos = lpos;
		if (lline[0] == '>' || lline[0] == ';') {
			strncpy(libstr,lline+1,20);
			if ((bp = strchr(libstr,' ')) != NULL) *bp = '\0';
			if ((bp = strchr(libstr,'\n')) != NULL) *bp = '\0';
			if ((bp = strchr(libstr,'\r')) != NULL) *bp = '\0';
			libstr[12]='\0';
			}
		else {
			libstr[0]='\0';
			strncpy((char *)seqp,lline,(size_t)(seqm-seqp));
			for (cp=seqp; seqp<seqm1; ) {
				if ((*seqp++=ap[*cp++])<NA) continue;
				if (*--seqp>NA) break;
				}
			if (*seqp==ES) goto done;
			}
#endif
		}

	lline[0] = '\0';
	while (seqp < seqm1 && my_fgets((char *)seqp, (size_t)(seqm-seqp),libf) != NULL) {
		if (*seqp == '>') goto new;
		if (*seqp == ';') {
			if (strchr((char *)seqp,'\n')==NULL &&
			    strchr((char *)seqp,'\r') == NULL) goto cont;
			continue;
			}
		for (cp=seqp; seqp<seqm1; ) {
			if ((*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA) continue;
			    if (*(--seqp)>NA) break;
			    }
		if (*seqp == ES) goto done;
		lpos = ftell(libf);
		}
	goto done;
new:	strncpy(lline,(char *)seqp,MAXLINE);
	lline[MAXLINE-1]='\0';
	if (strchr((char *)seqp,'\n') == NULL
            && strchr((char *)seqp,'\r') == NULL)
	  my_fgets(&lline[strlen(lline)],sizeof(lline)-strlen(lline),libf);
	goto done;

cont:
	my_fgets(lline,sizeof(lline),libf);
	seqm1 = seqp;

done:	if (seqp >= seqm1) {
		(*lcont)++;
		}
	else {
#ifdef NOLIB
	leof = 1;
#endif
	*lcont=0;
		}

	*seqp = EOSEQ;
	/*if ((long)(seqp - seq) == 0) return 1;*/
	return (long)(seqp - seq);
	}

/*=============================================================
*/
long RANLIB(char *str, long cnt, long seek)
{
  char *bp;
  long ll;

  fseek(libf, seek, 0);
  my_fgets(lline,sizeof(lline),libf);
  if (lline[0]=='>' || lline[0]==';') {
    strncpy(str,lline+1,cnt);
    str[cnt-1]='\0';
#ifdef SUPERFAMNUM
    if ((bp = strchr(str,SFCHAR))!=NULL) *bp='\0';
    else if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
    else if ((bp = strchr(str,'\r')) != NULL) *bp = '\0';
    else str[cnt-1]='\0';
#else
    if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
    else if ((bp = strchr(str,'\r')) != NULL) *bp = '\0';
    else str[cnt-1]='\0';
#endif
  }
  else {
    str[0]='\0';
  }
#ifdef NOLIB
  leof=0;
#endif
}

#ifndef NOLIB
char *cpsave;
/*=============================================================
*/
long lgetlib(char *seq, long maxs, char *libstr, long *libpos, long *lcont)
{
  long i, n, ll;
  long ic;
  register char *cp, *seqp;
  register long *ap;
  char *seqm, *seqm1;
  char *linep;

  seqp = seq;
  seqm = &seq[maxs-11];
  seqm1 = seqm-1;
#if defined(TFASTA) || defined(TFASTX)
	ap = nascii;
#else
	ap = sascii;
#endif
  if (*lcont==0) {
    while (lline[0]!='L' || lline[1]!='O' || 
	   strncmp(lline,"LOCUS",5)) { /* find LOCUS */
      lpos = ftell(libf);
      if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
      if (lfflag) getc(libf);
    }
    strncpy(libstr,&lline[12],13);
    libstr[12]='\0';
    *libpos=lpos;
    while (lline[0]!='O' || lline[1]!='R' ||
	   strncmp(lline,"ORIGIN",6)) { /* find ORIGIN */
      if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
      if (lfflag) getc(libf);
    }
  }
  else {
    for (cp= cpsave; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    }
  }

  lline[0]='\0';
  while (seqp<seqm1 && my_fgets(lline,sizeof(lline),libf)!=NULL) {
    if (lfflag) getc(libf);
    if (lline[0]=='/') goto new;
    for (cp= (char *)&lline[10]; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    }
  }
  goto done;
new:
  lpos = ftell(libf);
  my_fgets(lline,sizeof(lline),libf);
  if (lfflag) getc(libf);

done:
  if (seqp>=seqm1) {
    cpsave = cp;
    (*lcont)++;
  }
  else *lcont=0;

  *seqp = EOSEQ;
  if ((long)(seqp-seq)==0) return 1;
  return (long)(seqp-seq);
}
/*=============================================================
*/
long lranlib(char *str, long cnt, long seek)
{
  char *bp;
  long ll;

  fseek(libf, seek, 0);
  my_fgets(lline,sizeof(lline),libf);
  if (lfflag) getc(libf);

  strncpy(str,&lline[12],12);
  str[12]='\0';
  my_fgets(lline,sizeof(lline),libf);
  if (lfflag) getc(libf);
  while (lline[0]!='D' || lline[1]!='E' || strncmp(lline,"DEFINITION",10))
    my_fgets(lline,sizeof(lline),libf);
  strncpy(&str[10],&lline[11],cnt-10);
  str[cnt-1]='\0';
  if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';

  fseek(libf,seek,0);
  my_fgets(lline,sizeof(lline),libf);
  if (lfflag) getc(libf);
}
/*=============================================================
*/
long pgetlib(char *seq, long maxs, char *libstr, long *libpos, long *lcont)
{
	long i, n, ll;
	long ic;
	register char *cp, *seqp;
	register long *ap;
	char *seqm, *seqm1;
	char *linep;

	seqp = seq;
	seqm = &seq[maxs-11];
	seqm1 = seqm-1;
#if defined(TFASTA) || defined(TFASTX)
	ap = nascii;
#else
	ap = sascii;
#endif
	if (*lcont==0) {
		while (lline[0]!='E' || lline[1]!='N' || strncmp(lline,"ENTRY",5))
		{ /* find ENTRY */
			lpos = ftell(libf);
			if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
			}
		strncpy(libstr,&lline[16],8);
		libstr[8]='\0';
		*libpos = lpos;
		while (lline[0]!='S' || lline[2]!='Q' || strncmp(lline,"SEQUENCE",8))
		{ /* find SEQUENCE */
			if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
			}
		my_fgets(lline,sizeof(lline),libf); /* get the extra line */
		}
	else {
		for (cp= cpsave; seqp<seqm1; ) {
			if ((*seqp++=ap[*cp++])<NA) continue;
			if (*(--seqp)>NA) break;
			}
		if (*seqp==ES) goto done;
		}

	lline[0]='\0';
	while (seqp<seqm1 && my_fgets(lline,sizeof(lline),libf)!=NULL) {
		if (lline[0]=='/') goto new;
		for (cp= (char *)&lline[8]; seqp<seqm1; ) {
			if ((*seqp++=ap[*cp++])<NA) continue;
			if (*(--seqp)>NA) break;
		      };
		if (*seqp==ES) goto done;
		}
	goto done;
new:	lpos = ftell(libf);
	my_fgets(lline,sizeof(lline),libf);

done:	if (seqp>=seqm1) {
		cpsave = cp;
		(*lcont)++;
		}
	else *lcont=0;

	*seqp = EOSEQ;
	if ((long)(seqp-seq)==0) return 1;
	return (long)(seqp-seq);
	}
/*=============================================================
*/
long pranlib(char *str, long cnt, long seek)
{
	char *bp;
	long ll;

	fseek(libf, seek, 0);
	my_fgets(lline,sizeof(lline),libf);

	strncpy(str,&lline[16],8);
	str[8]='\0';
	my_fgets(lline,sizeof(lline),libf);
	while (lline[0]!='T' || lline[1]!='I' || strncmp(lline,"TITLE",5))
		my_fgets(lline,sizeof(lline),libf);
	strncpy(&str[8],&lline[16],cnt-9);
	str[cnt-1]='\0';
	if ((bp = strchr(str,'\n'))!=NULL &&
	    (bp = strchr(str,'\r'))!=NULL) *bp='\0';

	fseek(libf,seek,0);
	my_fgets(lline,sizeof(lline),libf);
}

long seqsiz;
/*=============================================================
*/
long egetlib(char *seq, long maxs, char *libstr, long *libpos, long *lcont)
{
	long ll;
	long ic;
	register char *cp, *seqp;
	register long *ap;
	char *seqm, *seqm1;
	char *linep;
	char id[11];  /* Holds Identifier */

	seqp = seq;
	seqm = &seq[maxs-11];
	seqm1 = seqm-1;
#if defined(TFASTA) || defined(TFASTX)
	ap = nascii;
#else
	ap = sascii;
#endif
	if (*lcont==0) {
		while (lline[0]!='I' || lline[1]!='D') { /* find ID */
			lpos = ftell(libf);
			if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
			if (lfflag) getc(libf);
			}
		sscanf(&lline[5],"%s",id);
		sprintf(libstr,"%-10.10s",id);
		*libpos = lpos;
		while (lline[0]!='S' || lline[1]!='Q') { /* find ORIGIN */
			if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
			if (lfflag) getc(libf);
			}
		sscanf(&lline[14],"%ld",&seqsiz);
		}
	else {
		for (cp= cpsave; seqp<seqm1; ) {
			if ((*seqp++=ap[*cp++])<NA) continue;
			if (*(--seqp)>NA) break;
			}
		if (*seqp==ES) goto done;
		}

	lline[0]='\0';
	while (seqp<seqm1 && my_fgets(lline,sizeof(lline),libf)!=NULL) {
		if (lfflag) getc(libf);
		if (lline[0]=='/') goto new;
		lline[70]='\0';
		for (cp= (char *)&lline[5]; seqp<seqm1; ) {
			if ((*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA) continue;
			    if (*(--seqp)>NA) break;
			    }
		if (*seqp==ES) goto done;
		}
	goto done;
new:	lpos = ftell(libf);
	my_fgets(lline,sizeof(lline),libf);
	if (lfflag) getc(libf);
	goto done;

done:	if (seqp>=seqm1) {
		cpsave = cp;
		(*lcont)++;
		seqsiz -= (long)(seqp-seq);
		}
	else *lcont=0;

	*seqp = EOSEQ;
	if ((long)(seqp-seq)==0) return 1;
/*	if (*lcont==0 && (long)(seqp-seq)!=seqsiz)
		printf("%s read %d of %d\n",libstr,(long)(seqp-seq),seqsiz);
*/
	return (long)(seqp-seq);
}
/*=============================================================
*/
long eranlib(char *str, long cnt, long seek)
{
	char *bp;
	char id[11];  /* Holds Identifier */
	long ll;

	fseek(libf, seek, 0);
	my_fgets(lline,sizeof(lline),libf);
	if (lfflag) getc(libf);

	sscanf(&lline[5],"%s",id);
	sprintf(str,"%-10.10s ",id);
	my_fgets(lline,sizeof(lline),libf);
	if (lfflag) getc(libf);
	while (lline[0]!='D' || lline[1]!='E') my_fgets(lline,sizeof(lline),libf);
	strncpy(&str[11],&lline[5],cnt-11);
	str[cnt-1]='\0';
	if ((bp = strchr(str,'\r'))!=NULL) *bp='\0';
	if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';

	fseek(libf,seek,0);
	my_fgets(lline,sizeof(lline),libf);
	if (lfflag) getc(libf);
}
/*=============================================================
*/
long igetlib(unsigned char *seq, long maxs, char *libstr, long *libpos, long *lcont)
{
	long i, n, ll;
	long ic;
	register unsigned char *cp, *seqp;
	register long *ap;
	unsigned char *seqm, *seqm1;
	char *linep, *bp;

	seqp = seq;
	seqm = &seq[maxs-9];
	seqm1 = seqm-1;
#if defined(TFASTA) || defined(TFASTX)
	ap = nascii;
#else
	ap = sascii;
#endif
	if (*lcont==0) {
		while (lline[0]!=';') {
			lpos = ftell(libf);
			if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
			}
		*libpos = lpos;
		while (lline[0]==';') my_fgets(lline,sizeof(lline),libf);
		strncpy(libstr,lline+1,10);
		libstr[9]='\0';
		if((bp=strchr(libstr,'\n'))!=NULL) *bp='\0';
		}

	lline[0]='\0';
	while (seqp<seqm1 && my_fgets((char *)seqp,(size_t)(seqm-seqp),libf)!=NULL) {
		if (*seqp=='>') goto new;
		if (*seqp==';') {
			if (strchr((char *)seqp,'\n')==NULL) goto cont;
			if (strchr((char *)seqp,'\r')==NULL) goto cont;
			continue;
			}
		for (cp=seqp; seqp<seqm1; ) {
			if ((*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA) continue;
			    if (*(--seqp)>NA) break;
			    }
		if (*seqp==ES) goto done;
		lpos = ftell(libf);
		}
	goto done;
new:	strncpy(lline,(char *)seqp,MAXLINE);
	lline[MAXLINE-1]='\0';
	if (strchr((char *)seqp,'\n')==NULL)
	    my_fgets(&lline[strlen(lline)],sizeof(lline)-strlen(lline),libf);
	goto done;

cont:
	my_fgets(lline,sizeof(lline),libf);
	seqm1 = seqp;

done:	if (seqp >= seqm1) {
		(*lcont)++;
		}
	else {
	*lcont=0;
		}


	*seqp = EOSEQ;
	if ((long)(seqp - seq) == 0) return 1;
	return (long)(seqp - seq);
}
/*=============================================================
*/
long iranlib(char *str, long cnt, long seek)
{
	char *bp;
	long ll;
	char tline[120];

	fseek(libf, seek, 0);
	my_fgets(lline,sizeof(lline),libf);

	if (lline[0]=='>' || lline[0]==';') {
		strncpy(tline,lline+1,sizeof(tline));
		str[cnt-1]='\0';
		if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
		if ((bp = strchr(str,'\r'))!=NULL) *bp='\0';
		else str[cnt-1]='\0';
		}
	else {
		tline[0]='\0';
		}

	while (lline[0]==';') my_fgets(lline,sizeof(lline),libf);
	if ((bp=strchr(lline,'\n'))!=NULL) *bp=0;
	if ((bp=strchr(lline,'\r'))!=NULL) *bp=0;
	if ((bp=strchr(lline,' '))!=NULL) *bp=0;
	strncpy(str,lline,cnt);
	strncat(str,"  ",cnt-strlen(str));
	strncat(str,tline,cnt-strlen(str));
	str[cnt-1]='\0';
	
	fseek(libf,seek,0);
	my_fgets(lline,sizeof(lline),libf);
	}
/*=============================================================
*/
long vgetlib(unsigned char *seq, long maxs, char *libstr, long *libpos, long *lcont)
{
	long i, n, ll;
	long ic, ich;
	register unsigned char *cp, *seqp;
	register long *ap;
	unsigned char *seqm, *seqm1;
	char *linep, *bp;

	seqp = seq;
	seqm = &seq[maxs-9];
	seqm1 = seqm-1;
#if defined(TFASTA) || defined(TFASTX)
	ap = nascii;
#else
	ap = sascii;
#endif
	if (*lcont==0) {
		while (lline[0]!='>' && lline[0]!=';') {
			lpos = ftell(libf);
			if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
			if (lfflag) getc(libf);
		}
		if ((bp=strchr(lline,SFCHAR))!=NULL) {
			*bp='\0';
			sscanf(bp+1,"%d",&sfnum);
		      }
		else sfnum=0;
		if ((bp=strchr(lline,'\n'))!=NULL) *bp='\0';
		if ((bp=strchr(lline,'\r'))!=NULL) *bp='\0';
		strncpy(libstr,&lline[4],20);
		if ((bp=strchr(libstr,' '))!=NULL) *bp = '\0';
		if ((bp=strchr(libstr,'\n'))!=NULL) *bp = '\0';
		if ((bp=strchr(libstr,'\r'))!=NULL) *bp = '\0';
		libstr[12]='\0';

		my_fgets(lline,sizeof(lline),libf);
		if (lfflag) getc(libf);
		*libpos = lpos;
		}

	lline[0]='\0';
	while (seqp<seqm1 && my_fgets((char *)seqp,(size_t)(seqm-seqp),libf)!=NULL) {
		if (lfflag && (ich=getc(libf))!=LFCHAR) ungetc(ich,libf);
		if (*seqp=='>') goto new;
		if (*seqp==';') {
			if (strchr((char *)seqp,'\n')==NULL) goto cont;
			continue;
			}
		for (cp=seqp; seqp<seqm1; ) {
			if ((*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA) continue;
			    if (*(--seqp)>NA) break;
			    }
		if (*seqp==ES) goto done;
		lpos = ftell(libf);
		}
	goto done;
new:	strncpy(lline,(char *)seqp,MAXLINE);
	lline[MAXLINE-1]='\0';
	if (strchr((char *)seqp,'\n')==NULL && 
	    strchr((char *)seqp,'\r')==NULL) {
		my_fgets(lline,sizeof(lline)-strlen(lline),libf);
		if (lfflag && (ich=getc(libf))!=LFCHAR) ungetc(ich,libf);
	}
	goto done;

cont:
	my_fgets(lline,sizeof(lline),libf);
	if (lfflag && (ich=getc(libf))!=LFCHAR) ungetc(ich,libf);
	seqm1 = seqp;

done:	if (seqp>=seqm1) {
		(*lcont)++;
		}
	else {

	*lcont=0;
		}


	*seqp = EOSEQ;
	if ((long)(seqp-seq)==0) return 1;
	return (long)(seqp-seq);
}
/*=============================================================
*/
long vranlib(char *str, long cnt, long seek)
{
  char *bp;
  long ll;

  fseek(libf, seek, 0);
  my_fgets(lline,sizeof(lline),libf);
  if (lfflag) getc(libf);

  if (lline[0]=='>'&&(lline[3]==';'||lline[3]=='>')) {
    strncpy(str,&lline[4],cnt);

    if ((bp = strchr(str,':'))!=NULL) *bp='\0';
    if ((bp=strchr(str,'\r'))!=NULL) *bp='\0';
    else if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
    else str[cnt-1]='\0';

    my_fgets(lline,sizeof(lline),libf);
    if (lfflag) getc(libf);

    if ((bp=strchr(lline,'\r'))!=NULL) *bp=' ';
    if ((bp=strchr(lline,'\n'))!=NULL) *bp='\0';
    strncat(str," ",(size_t)cnt);
    strncat(str,lline,(size_t)cnt-strlen(str));
  }
  else {
    str[0]='\0';
  }

  fseek(libf,seek,0);
  my_fgets(lline,sizeof(lline),libf);
  if (lfflag) getc(libf);
}

static char gcg_type[10];
static long gcg_len;
static long gcg_bton[4]={1,3,0,2};
/*=============================================================
*/
long gcg_getlib(unsigned char *seq, long maxs, char *libstr, long *libpos, long *lcont)
{
  char dummy[20];
  char gcg_date[6];
  long i, n, ll;
  long ic, ich;
  register unsigned char *cp, *seqp, stmp;
  register long *ap;
  unsigned char *seqm, *seqm1;
  long r_block, b_block;
  char *linep, *bp;

  seqp = seq;
  seqm = &seq[maxs-9];
  seqm1 = seqm-1;
#if defined(TFASTA) || defined(TFASTX)
	ap = nascii;
#else
	ap = sascii;
#endif
  if (*lcont==0) {
    while (lline[0]!='>' && lline[0]!=';') {
      lpos = ftell(libf);
      if (my_fgets(lline,sizeof(lline),libf)==NULL) return 0;
    }
    sscanf(&lline[4],"%s %s %s %s %ld",
	   libstr,gcg_date,gcg_type,dummy,&gcg_len);
    my_fgets(lline,sizeof(lline),libf);
    libstr[12]='\0';
    *libpos = lpos;
  }

  lline[0]='\0';

  r_block = b_block = min((size_t)(seqm-seqp),gcg_len);
  if (gcg_type[0]=='2') {
    r_block = (r_block+3)/4;
  }

  fread((char *)seqp,(size_t)r_block,(size_t)1,libf);
  if (gcg_type[0]=='A') 
    for (cp=seqp; seqp<seq+r_block; ) *seqp++ = ap[*cp++];
  else if (gcg_type[0]=='2') {
    seqp = seq + r_block;
    cp = seq + 4*r_block;
    while (seqp > seq) {
      stmp = *--seqp;
      *--cp = gcg_bton[stmp&3];
      *--cp = gcg_bton[(stmp >>= 2)&3];
      *--cp = gcg_bton[(stmp >>= 2)&3];
      *--cp = gcg_bton[(stmp >>= 2)&3];
    }
  }
  if (b_block == gcg_len) {
    my_fgets(lline,MAXLINE,libf);
    *lcont = 0;
  }
  else {
    if (gcg_type[0]=='2') b_block = 4*r_block;
    gcg_len -= b_block;
    (*lcont)++;
  }

  seq[b_block] = EOSEQ;
  if (b_block==0) return 1;
  else return b_block;
}
/*=============================================================
*/
long gcg_ranlib(char *str, long cnt, long seek)
{
  char *bp, *bp1, *bp2, *llp;
  long ll;

  fseek(libf, seek, 0);
  my_fgets(lline,sizeof(lline),libf);

  if (lline[0]=='>'&&(lline[3]==';'||lline[3]=='>')) {
    strncpy(str,&lline[4],cnt);

    if ((bp = strchr(str,':'))!=NULL) *bp='\0';
    if ((bp = strchr(str,' '))!=NULL) *bp='\0';
    if ((bp=strchr(str,'\r'))!=NULL) *bp='\0';
    else if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
    else str[cnt-1]='\0';

    my_fgets(lline,sizeof(lline),libf);

    for (llp=lline,bp=str; *llp == *bp; llp++, bp++);
    if ((long)(llp-lline)<5) llp=lline;

    /* here we would like to skip over some species stuff */
    if ((bp1 = strchr(llp,';'))!=NULL && (long)(bp1-llp)<50) {
      if ((bp2 = strchr(bp1+1,';'))!=NULL && (long)(bp2-bp1)<50) {
	*(bp2+1)='\0'; bp1 = bp2+2;
      }
      else {bp1=llp;}
    }
    else if ((bp1=strchr(llp,'.'))!=NULL && *(bp1+1)==' ') {
      *(bp1+1) = '\0'; bp1 += 2;
    }
    else bp1 = llp;

    if ((bp=strchr(bp1,'\r'))!=NULL) *bp='\0';
    if ((bp=strchr(bp1,'\n'))!=NULL) *bp='\0';
    strncat(str," ",(size_t)cnt);
    strncat(str,bp1,(size_t)cnt-strlen(str));
    if (bp1!=llp) strncat(str,llp,(size_t)cnt-strlen(str));
  }
  else {
    str[0]='\0';
  }

  fseek(libf,seek,0);
  my_fgets(lline,sizeof(lline),libf);
}

#endif	/* NOLIB */

/*=============================================================
*/
scanseq(char *seq, long n, char *str)
{
	long tot,i;
	char aaray[MAXSQ];		/* this must be set > nsq */
	
	for (i=0; i<MAXSQ; i++)  aaray[i]=0;
	for (i=0; i<strlen(str); i++) aaray[sascii[str[i]]]=1;
	for (i=tot=0; i<n; i++) tot += aaray[seq[i]];
	return tot;
}
/*=============================================================
*/
void revcomp(char *seq, long n)
{
  char tmp;
  long i, ni;

  for (i=0; i< n; i++)
    if (sq[seq[i]]=='A') seq[i] = nascii['T'];
    else if (sq[seq[i]]=='C') seq[i] = sascii['G'];
    else if (sq[seq[i]]=='G') seq[i] = sascii['C'];
    else if (sq[seq[i]]=='T') seq[i] = sascii['A'];
    else if (sq[seq[i]]=='R') seq[i] = sascii['Y'];
    else if (sq[seq[i]]=='Y') seq[i] = sascii['R'];
    else if (sq[seq[i]]=='M') seq[i] = sascii['K'];
    else if (sq[seq[i]]=='K') seq[i] = sascii['M'];
    else if (sq[seq[i]]=='D') seq[i] = sascii['H'];
    else if (sq[seq[i]]=='H') seq[i] = sascii['D'];
    else if (sq[seq[i]]=='V') seq[i] = sascii['B'];
    else if (sq[seq[i]]=='B') seq[i] = sascii['V'];

  for (i=0, ni = n-1; i< n/2; i++,ni--) {
    tmp = seq[i];
    seq[i] = seq[ni];
    seq[ni] = tmp;
  }
}
   
